__version_info__ = (1, 2, 1)
__version__ = '.'.join(map(str, __version_info__))
__author__ = "cs91chris <cs91chris@voidbrain.me>"

__all__ = [
    '__version_info__',
    '__version__',
    '__author__'
]



from flask import Blueprint

from flask_json import as_json

from sqlalchemy.ext.automap import automap_base
from sqlalchemy.ext.declarative import declarative_base

from flask_autocrud.model import Model
from flask_autocrud.service import Service

from .config import MESSAGES
from .config import set_default_config


class AutoCrud(object):
    def __init__(self, app=None, db=None, admin=None, view=None,
                 exclude_tables=None, user_models=None, schema=None):
        """

        :param db:
        :param app:
        :param admin:
        :param exclude_tables:
        :param user_models:
        :param schema:
        """
        self._app = app
        self._db = db
        self._admin = admin
        self._view = view
        self._api = None
        self._automap_model = None
        self._exclude_tables = exclude_tables
        self._user_models = user_models
        self._schema = schema
        self._models = {}

        if app is not None:
            self.init_app(self._app, self._db, self._admin, self._view)

    def init_app(self, app, db, admin=None, view=None,
                 exclude_tables=None, user_models=None, schema=None):
        """

        :param app:
        :param db:
        :param admin:
        :param view:
        :param exclude_tables:
        :param user_models:
        :param schema:
        :return:
        """
        self._app = app
        self._db = db
        self._admin = admin
        self._view = view
        self._exclude_tables = exclude_tables
        self._user_models = user_models
        self._schema = schema

        if self._db is None:
            raise AttributeError(MESSAGES.get('DB_NOT_NULL'))

        if self._admin and not self._view:
            raise AttributeError(MESSAGES.get('VIEW_NOT_NULL'))

        self._automap_model = automap_base(declarative_base(cls=(db.Model, Model)))
        set_default_config(self._app)

        self._api = Blueprint(
            'flask_autocrud',
            __name__,
            subdomain=self._app.config['AUTOCRUD_SUBDOMAIN']
        )

        with self._app.app_context():
            if self._user_models:
                if any([issubclass(cls, self._automap_model) for cls in self._user_models]):
                    self._automap_model.prepare(self._db.engine, reflect=True, schema=self._schema)

                for user_model in self._user_models:
                    self._register_model(user_model)
            else:
                self._automap_model.prepare(self._db.engine, reflect=True, schema=self._schema)

                for cls in self._automap_model.classes:
                    if self._exclude_tables and cls.__table__.name in self._exclude_tables:
                        continue

                    if self._app.config['AUTOCRUD_READ_ONLY']:
                        cls.__methods__ = {'GET', 'FETCH'}

                    if not self._app.config['AUTOCRUD_FETCH_ENABLED']:
                        cls.__methods__ -= {'FETCH'}

                    self._register_model(cls)

        if self._app.config['AUTOCRUD_RESOURCES_URL_ENABLED']:
            @self._api.route(self._app.config['AUTOCRUD_BASE_URL'] + self._app.config['AUTOCRUD_RESOURCES_URL'])
            @as_json
            def index():
                """

                :return:
                """
                routes = {}
                for name, cls in self._models.items():
                    routes[name] = "{}{{/{}}}".format(cls.__url__, cls.primary_key())
                return routes

        self._app.register_blueprint(self._api)

        if not hasattr(app, 'extensions'):
            app.extensions = dict()
        app.extensions['autocrud'] = self

    def _register_model(self, cls):
        """

        :param cls:
        """
        if self._admin is not None:
            self._admin.add_view(self._view(cls, self._db.session))

        class_name = cls.__name__
        cls.__url__ = self._app.config['AUTOCRUD_BASE_URL'] + '/' + class_name.lower()

        service_class = type(
            class_name + 'Service',
            (Service,),
            {
                '__model__': cls,
                '__db__': self._db,
                '__collection_name__': class_name
            }
        )

        model_url = cls.__url__
        methods = set(cls.__methods__)
        self._models.update({cls.__name__: cls})
        view_func = service_class.as_view(class_name.lower())

        if 'GET' in methods:
            self._api.add_url_rule(
                model_url,
                defaults={'resource_id': None},
                view_func=view_func,
                strict_slashes=False,
                methods=['GET']
            )

            if self._app.config['AUTOCRUD_METADATA_ENABLED'] is True:
                self._api.add_url_rule(
                    '{resource}{meta}'.format(
                        resource=model_url,
                        meta=self._app.config['AUTOCRUD_METADATA_URL']
                    ),
                    view_func=view_func,
                    methods=['GET']
                )

        if 'POST' in methods:
            self._api.add_url_rule(model_url, view_func=view_func, methods=['POST'])

        if 'FETCH' in methods:
            self._api.add_url_rule(model_url, view_func=view_func, methods=['FETCH'])

        cols = list(cls().__table__.primary_key.columns)

        self._api.add_url_rule(
            '{resource}/<{pk_type}:{pk}>'.format(
                resource=model_url,
                pk='resource_id',
                pk_type=cols[0].type.python_type.__name__ if len(cols) > 0 else 'string'
            ),
            view_func=view_func,
            strict_slashes=False,
            methods=methods - {'POST', 'FETCH'}
        )



import io
import csv

from flask import Response
from flask import make_response

from flask_json import as_json

from .config import HTTP_STATUS
from .model import Model


def list_to_csv(data: list, delimiter=';', quoting=True, qc='"'):
    """

    :param data:
    :param delimiter:
    :param quoting:
    :param qc:
    :return:
    """
    q = csv.QUOTE_ALL if quoting else csv.QUOTE_NONE
    output = io.StringIO()

    w = csv.DictWriter(
        output,
        data[0].keys() if data else '',
        dialect='excel-tab',
        delimiter=delimiter,
        quotechar=qc,
        quoting=q
    )
    w.writeheader()
    w.writerows(data)

    return output.getvalue()


def resp_csv(data_list, filename):
    """

    :param data_list:
    :param filename:
    :return:
    """
    return Response(
        list_to_csv(data_list),
        mimetype='text/csv',
        headers={
            'Content-Type': 'text/csv',
            'Total-Count': len(data_list),
            'Content-Disposition': 'attachment; filename=%s.csv' % (filename,)
        }
    )


@as_json
def resp_json(data, root=None, code=HTTP_STATUS.SUCCESS):
    """

    :param data:
    :param root:
    :param code:
    :return:
    """
    if root is not None:
        return {
            str(root): data
        }, code
    else:
        return data, code


@as_json
def response_with_links(resource, code=HTTP_STATUS.SUCCESS):
    """

    :param resource:
    :param code:
    :return:
    """
    links = resource.links()
    link_string = '<{}>; rel=self'.format(links['self'])

    for link in links.values():
        link_string += ', <{}>; rel=related'.format(link)

    return resource.to_dict(), code, {
        'Link': link_string
    }


@as_json
def response_with_location(resource, code=HTTP_STATUS.CREATED):
    """

    :param resource:
    :param code:
    :return:
    """
    location = resource.links()

    return resource.to_dict(), code, {
        'Location': location['self']
    }


def no_content():
    """

    :return:
    """
    resp = make_response('', HTTP_STATUS.NO_CONTENT)
    del resp.headers['Content-Type']
    del resp.headers['Content-Length']
    return resp


@as_json
def response_with_pagination(resource, pagination):
    """

    :param resource:
    :param pagination:
    :return:
    """
    code = HTTP_STATUS.SUCCESS

    total_results = pagination.total_results
    page_number = pagination.page_number
    num_pages = pagination.num_pages
    page_size = pagination.page_size

    if num_pages > 1 and total_results > page_size:
        code = HTTP_STATUS.PARTIAL_CONTENT

    if page_number == num_pages:
        code = HTTP_STATUS.SUCCESS

    if isinstance(resource, Model):
        resource = resource.to_dict()

    return resource, code, {
        'Pagination-Count': total_results,
        'Pagination-Page': page_number,
        'Pagination-Num-Pages': num_pages,
        'Pagination-Limit': page_size
    }



from datetime import datetime
from decimal import Decimal

from collections import MutableMapping

from .model import Model
from .validators import valid_number

from .config import ARGUMENT
from .config import COLLECTION_SUFFIX


def get_pagination_params(conf, args):
    """

    :param conf:
    :param args:
    :return:
    """
    invalid = []
    page = args.get(ARGUMENT.STATIC.page)
    limit = args.get(ARGUMENT.STATIC.limit)

    page = valid_number(page)
    limit = valid_number(limit)

    invalid.append(ARGUMENT.STATIC.page) if page is False else None
    invalid.append(ARGUMENT.STATIC.limit) if limit is False else None

    if conf.get('AUTOCRUD_QUERY_LIMIT_ENABLED'):
        page = 1 if not page else page
        max_limit = conf.get('AUTOCRUD_MAX_QUERY_LIMIT')

        if not limit or limit > max_limit:
            limit = conf.get('AUTOCRUD_MAX_QUERY_LIMIT')

    return page, limit, invalid


def from_model_to_dict(data):
    """

    :param data:
    :return:
    """
    resp = {}

    for k, v in data.items():
        if k.startswith('_'):
            continue

        if isinstance(v, Model):
            resp.update({
                v.__class__.__name__: from_model_to_dict(v.__dict__)
            })
        elif isinstance(v, list):
            if len(v) > 0:
                name = v[0].__class__.__name__ + COLLECTION_SUFFIX
                resp.update({
                    name: [from_model_to_dict(i.__dict__) for i in v]
                })
        else:
            if isinstance(v, Decimal):
                v = float(v)
            elif isinstance(v, datetime):
                v = v.isoformat()

            resp.update({k: v})
    return resp


def to_flatten_dict(d, parent_key='', sep='_'):
    """

    :param d:
    :param parent_key:
    :param sep:
    :return:
    """
    items = []

    for k, v in d.items():
        new_key = (parent_key + sep + k) if parent_key else k

        if isinstance(v, MutableMapping):
            items.extend(to_flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)



from flask import request
from flask import current_app as cap
from flask.views import MethodView

from sqlalchemy import inspect
from sqlalchemy.exc import ArgumentError
from sqlalchemy.orm import contains_eager

from sqlalchemy_filters import apply_filters
from sqlalchemy_filters import apply_loads
from sqlalchemy_filters import apply_sort
from sqlalchemy_filters import apply_pagination

from sqlalchemy_filters.exceptions import BadSpec
from sqlalchemy_filters.exceptions import FieldNotFound
from sqlalchemy_filters.exceptions import BadFilterFormat
from sqlalchemy_filters.exceptions import BadSortFormat

from .wrapper import resp_csv
from .wrapper import resp_json
from .wrapper import no_content
from .wrapper import response_with_links
from .wrapper import response_with_location
from .wrapper import response_with_pagination

from .validators import validate_entity
from .validators import parsing_query_string

from .utils import get_pagination_params
from .utils import from_model_to_dict
from .utils import to_flatten_dict

from .config import ARGUMENT
from .config import GRAMMAR
from .config import HTTP_STATUS
from .config import COLLECTION_SUFFIX


class Service(MethodView):
    __db__ = None
    __model__ = None
    __collection_name__ = 'resources'

    def delete(self, resource_id):
        """

        :param resource_id:
        :return:
        """
        model = self.__model__
        session = self.__db__.session()

        resource = model.query.get(resource_id)
        if not resource:
            return resp_json({'message': 'Not Found'}, code=HTTP_STATUS.NOT_FOUND)

        session.delete(resource)
        session.commit()
        return no_content()

    def get(self, resource_id=None):
        """

        :param resource_id:
        :return:
        """
        response = []
        invalid = []
        model = self.__model__
        export = True if ARGUMENT.STATIC.export in request.args else False
        extended = True if ARGUMENT.STATIC.extended in request.args else False

        if resource_id is not None:
            resource = model.query.get(resource_id)
            if not resource:
                return resp_json({'message': 'Not Found'}, code=HTTP_STATUS.NOT_FOUND)
            return response_with_links(resource)

        if request.path.endswith('meta'):
            return resp_json(model.description())

        page, limit, error = get_pagination_params(cap.config, request.args)
        invalid += error

        fields, statement, error = parsing_query_string(model)
        invalid += error

        if len(invalid) > 0:
            return resp_json(invalid, 'invalid', code=HTTP_STATUS.BAD_REQUEST)

        statement, pagination = apply_pagination(statement, page, limit)
        resources = statement.all()

        for r in resources:
            item = r.to_dict(True if extended else False)
            item_keys = item.keys()

            if fields:
                for k in set(item_keys) - set(fields):
                    item.pop(k)

            if export and extended:
                item = to_flatten_dict(item)

            response.append(item)

        if export:
            file_name = self.__collection_name__
            file_name += ("_" + str(page)) if page else ""
            file_name += ("_" + str(limit)) if limit else ""
            return resp_csv(response, file_name)

        return response_with_pagination(response, pagination)

    def patch(self, resource_id):
        """

        :param resource_id:
        :return:
        """
        model = self.__model__
        session = self.__db__.session()

        data = request.get_json() or {}
        validate_entity(model, data)

        resource = model.query.get(resource_id)
        if not resource:
            return resp_json({'message': 'Not Found'}, code=HTTP_STATUS.NOT_FOUND)

        resource.update(data)
        session.merge(resource)
        session.commit()

        return response_with_links(resource)

    def post(self):
        """

        :return:
        """
        model = self.__model__
        session = self.__db__.session()

        data = request.get_json()
        if not data:
            return resp_json({'message': 'Bad Request'}, code=HTTP_STATUS.BAD_REQUEST)

        validate_entity(model, data)

        resource = model.query.filter_by(**data).first()
        if not resource:
            resource = model(**data)
            session.add(resource)
            session.commit()
            code = HTTP_STATUS.CREATED
        else:
            code = HTTP_STATUS.CONFLICT

        return response_with_location(resource, code)

    def put(self, resource_id):
        """

        :param resource_id:
        :return:
        """
        model = self.__model__
        session = self.__db__.session()

        data = request.get_json() or {}
        validate_entity(model, data)

        resource = model.query.get(resource_id)
        if resource:
            resource.update(data)
            session.merge(resource)
            session.commit()

            return response_with_links(resource)

        resource = model(**data)
        session.add(resource)
        session.commit()

        return response_with_links(resource, HTTP_STATUS.CREATED)

    def fetch(self):
        """

        :return:
        """
        response = []
        invalid = []
        model = self.__model__
        query = self.__db__.session.query(self.__model__)

        data = request.get_json() or {}
        joins = data.get('joins') or {}
        filters = data.get('filters') or []
        fields = data.get('fields') or []
        sort = data.get('sortBy') or []
        export = True if ARGUMENT.STATIC.export in request.args else False
        page, limit, error = get_pagination_params(cap.config, data.get('pagination') or {})
        invalid += error

        cap.logger.debug(query)

        for k in fields:
            if k not in (model.required() + model.optional()):
                invalid.append(k)

        if len(invalid) == 0 and len(fields) > 0:
            query = apply_loads(query, fields)
            cap.logger.debug(query)

        for k in joins.keys():
            instance = None
            for r in inspect(model).relationships:
                if r.key == k.lower() or r.key.split('_collection')[0] == k.lower():
                    instance = getattr(model, r.key)

            if instance is not None:
                joint = joins.get(k)
                try:
                    load_column = contains_eager(instance)
                    if len(joint) > 0 and joint[0] != GRAMMAR.ALL:
                        load_column = load_column.load_only(*joint)

                    query = query.join(instance, aliased=False).options(load_column)
                    cap.logger.debug(query)
                except ArgumentError:
                    invalid += joint
            else:
                invalid.append(k)

        for f in filters:
            try:
                query = apply_filters(query, f)
                cap.logger.debug(query)
            except BadSpec:
                invalid.append(f.get('model'))
            except FieldNotFound:
                invalid.append(f.get('field'))
            except BadFilterFormat:
                invalid.append(f.get('op'))

        for s in sort:
            try:
                query = apply_sort(query, s)
                cap.logger.debug(query)
            except BadSpec:
                invalid.append(s.get('model'))
            except FieldNotFound:
                invalid.append(s.get('field'))
            except BadSortFormat:
                invalid.append(s.get('direction'))

        if len(invalid) > 0:
            return resp_json(invalid, 'invalid', code=HTTP_STATUS.BAD_REQUEST)

        query, pagination = apply_pagination(query, page, limit)

        for r in query.all():
            zipkeys = {}
            data = from_model_to_dict(r.__dict__)

            if export:
                data = to_flatten_dict(data)

                for key in list(data.keys()):
                    if isinstance(data.get(key), list):
                        zipkeys.update({key.rstrip(COLLECTION_SUFFIX) + '_': data.get(key)})
                        del data[key]

                for zk, value in zipkeys.items():
                    response.append({
                        **data,
                        **{zk+k: v for k, v in value[0].items()}
                    })

            if len(zipkeys.keys()) == 0:
                response.append(data)

        if export:
            file_name = self.__collection_name__
            file_name += ("_" + str(page)) if page else ""
            file_name += ("_" + str(limit)) if limit else ""
            return resp_csv(response, file_name)

        return response_with_pagination(response, pagination)



"""
Flask-AutoCRUD
-------------

Automatically generate a RESTful API service for CRUD operation on database
"""

from setuptools import setup

from flask_autocrud import __version__
from flask_autocrud import __author__

author, email = __author__.split()
email = email.lstrip('<').rstrip('>')

with open("README.rst", "r") as fh:
    long_description = fh.read()


setup(
    name='Flask-AutoCRUD',
    version=__version__,
    url='https://github.com/cs91chris/flask_autocrud/',
    license='MIT',
    author=author,
    author_email=email,
    description='Automatically generate a RESTful API service for CRUD operation on database',
    long_description=long_description,
    packages=['flask_autocrud'],
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    install_requires=[
        'Flask==1.0.2',
        'Flask-Admin==1.5.3',
        'Flask-SQLAlchemy==2.3.2',
        'Flask-JSON==0.3.3',
        'sqlalchemy-filters==0.10.0'
    ],
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



class GRAMMAR:
    SEP = ';'
    NOT = '!'
    REVERSE = '-'
    NULL = 'null'
    LIKE = '%'
    NOT_NULL = NOT + NULL
    NOT_LIKE = NOT + LIKE
    ALL = '*'


class ARGUMENT:
    class STATIC:
        page = 'page'
        limit = 'limit'
        export = 'export'
        extended = 'extended'

    class DYNAMIC:
        sort = 'sort'
        fields = 'fields'


class HTTP_STATUS:
    SUCCESS = 200
    CREATED = 201
    NO_CONTENT = 204
    PARTIAL_CONTENT = 206
    BAD_REQUEST = 400
    NOT_FOUND = 404
    CONFLICT = 409
    UNPROCESSABLE_ENTITY = 422
    INTERNAL_SERVER_ERROR = 500


ALLOWED_METHODS = {
    'GET',
    'POST',
    'PUT',
    'PATCH',
    'DELETE',
    'FETCH',
    'HEAD',
    'OPTIONS'
}

MODEL_VERSION = '1'
COLLECTION_SUFFIX = 'List'


MESSAGES = {
    'DB_NOT_NULL': """
        You can not create AutoCrud without an SQLAlchemy instance.
        Please consider to use the init_app method instead
    """,
    'VIEW_NOT_NULL': """
        You can not create AutoCrud with Admin but without ModelView instance.
        admin and view arguments are required together
    """
}


def set_default_config(app):
    """

    :param app:
    """
    app.config.setdefault('AUTOCRUD_METADATA_ENABLED', True)
    app.config.setdefault('AUTOCRUD_METADATA_URL', '/meta')
    app.config.setdefault('AUTOCRUD_READ_ONLY', False)
    app.config.setdefault('AUTOCRUD_BASE_URL', '')
    app.config.setdefault('AUTOCRUD_RESOURCES_URL', '/resources')
    app.config.setdefault('AUTOCRUD_SUBDOMAIN', None)
    app.config.setdefault('AUTOCRUD_RESOURCES_URL_ENABLED', True)
    app.config.setdefault('AUTOCRUD_QUERY_LIMIT_ENABLED', True)
    app.config.setdefault('AUTOCRUD_MAX_QUERY_LIMIT', 1000)
    app.config.setdefault('AUTOCRUD_FETCH_ENABLED', True)
